import os
import random
import unittest
import shutil, tempfile

import numpy as np

import stk

class Test_Volume(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_constructor(self):
        arr = np.zeros((4, 5, 6)).astype(np.int32)
        vol = stk.Volume(arr)

        np.testing.assert_equal(vol.size, (6, 5, 4))
        np.testing.assert_equal(vol.origin, (0, 0, 0))
        np.testing.assert_equal(vol.spacing, (1, 1, 1))
        np.testing.assert_equal(vol.direction, ((1, 0, 0), (0, 1, 0), (0, 0, 1)))
        np.testing.assert_equal(np.array(vol), arr)

        d = ((10, 0, 0), (0, 11, 0), (0, 0, 12))
        vol = stk.Volume(arr, origin=(1,2,3), spacing=(4,5,6), direction=d)

        np.testing.assert_equal(vol.size, (6, 5, 4))
        np.testing.assert_equal(vol.origin, (1, 2, 3))
        np.testing.assert_equal(vol.spacing, (4, 5, 6))
        np.testing.assert_equal(vol.direction, d)
        np.testing.assert_equal(np.array(vol), arr)

    def test_meta(self):
        vol = stk.Volume()
        vol.origin = (2, 3, 4)
        np.testing.assert_equal(vol.origin, (2, 3, 4))

        vol = stk.Volume()
        vol.spacing = (2, 3, 4)
        np.testing.assert_equal(vol.spacing, (2, 3, 4))

        d = ((10, 0, 0), (0, 11, 0), (0, 0, 12))
        vol = stk.Volume()
        vol.direction = d
        np.testing.assert_equal(vol.direction, d)

    def test_copy_meta_from(self):
        vol1 = stk.Volume()
        vol2 = stk.Volume()

        vol1.origin = (2,3,4)
        vol1.spacing = (5,6,7)
        d = ((10, 0, 0), (0, 11, 0), (0, 0, 12))
        vol1.direction = d

        vol2.copy_meta_from(vol1)

        np.testing.assert_equal(vol2.origin, (2, 3, 4))
        np.testing.assert_equal(vol2.spacing, (5, 6, 7))
        np.testing.assert_equal(vol2.direction, d)
    
    def test_type(self):
        data = np.zeros((5,5,5)).astype(np.float32)
        vol = stk.Volume(data)
        self.assertEqual(vol.type, stk.Type.Float)

        data = np.zeros((5,5,5,3)).astype(np.float32)
        vol = stk.Volume(data)
        self.assertEqual(vol.type, stk.Type.Float3)

    def test_io(self):
        data = np.zeros((5,5,5)).astype(np.float32)
        vol1 = stk.Volume(data, origin=(2,3,4), spacing=(5,6,7))

        f = os.path.join(self.temp_dir, 'test.nrrd')
        stk.write_volume(f, vol1)

        vol2 = stk.read_volume(f)
        np.testing.assert_equal(vol2.origin, vol1.origin)
        np.testing.assert_equal(vol2.spacing, vol1.spacing)
        np.testing.assert_equal(vol2.direction, vol1.direction)
        np.testing.assert_equal(np.array(vol2), np.array(vol1))



import os
import random
import unittest

import numpy as np
import SimpleITK as sitk

from random import uniform
from numpy.random import rand, randint

import pydeform.sitk_api as pydeform
import _stk as stk

# Use a known, random seed for each assert when
# testing with random data.
def _set_seed():
    seed = int.from_bytes(os.urandom(4), byteorder="big")
    np.random.seed(seed)
    random.seed(seed)
    return seed


def _gauss3(size=(200, 200, 200), mu=(100, 100, 100), sigma=20, gamma=1):
    x = np.linspace(0, size[2], size[2])
    y = np.linspace(0, size[1], size[1])
    z = np.linspace(0, size[0], size[0])
    x, y, z = np.meshgrid(x, y, z, indexing='ij', sparse=True)
    arr = gamma * np.exp(-((x-mu[2])/sigma)**2 - ((y-mu[1])/sigma)**2 - ((z-mu[0])/sigma)**2)
    return sitk.GetImageFromArray(arr.astype(np.float32))


def _show(image, origin=(0, 0, 0), spacing=(1, 1, 1), direction=(1, 0, 0, 0, 1, 0, 0, 0, 1)):
    image = sitk.GetImageFromArray(image)
    image.SetOrigin(origin)
    image.SetSpacing(spacing)
    image.SetDirection(direction)
    sitk.Show(image)


def _jaccard(a, b):
    a = sitk.GetArrayFromImage(a)
    b = sitk.GetArrayFromImage(b)
    return np.sum(np.logical_and(a, b)) / np.sum(np.logical_or(a, b))

class Test_SitkAPI(unittest.TestCase):

    def test_register(self):

        with self.assertRaises(RuntimeError):
            pydeform.register(None, None)

        fixed = _gauss3((10, 10, 10))
        with self.assertRaises(RuntimeError):
            pydeform.register(fixed, None)

        moving = _gauss3((10, 10, 10))
        with self.assertRaises(RuntimeError):
            pydeform.register(None, moving)

        fixed = _gauss3((10, 10, 10))
        moving = [fixed, fixed]
        with self.assertRaises(ValueError):
            pydeform.register(fixed, moving)

        moving = _gauss3((10, 10, 10))
        fixed = [moving, moving]
        with self.assertRaises(ValueError):
            pydeform.register(fixed, moving)

        fixed = sitk.Cast(_gauss3(size=(40, 50, 60), mu=(20, 25, 30), sigma=8) > 0.3,
                          sitk.sitkFloat32)
        moving = sitk.Cast(_gauss3(size=(40, 50, 60), mu=(30, 20, 25), sigma=8) > 0.3,
                           sitk.sitkFloat32)

        settings = {
            'regularization_weight': 0.05,
        }

        d = pydeform.register(fixed, moving, settings=settings)

        res = pydeform.transform(moving, d, sitk.sitkNearestNeighbor)

        self.assertGreater(_jaccard(res > 0.1, fixed > 0.1), 0.97)

    def test_jacobian(self):
        for _ in range(100):
            seed = _set_seed()

            # Generate some random image data
            pad = 1
            origin = [uniform(-5, 5) for i in range(3)]
            spacing = [uniform(0.1, 5) for i in range(3)]
            shape_no_pad = [randint(50, 80) for i in range(3)]
            d = 5 * (2.0 * rand(*shape_no_pad, 3) - 1.0)
            d = np.pad(d, 3 * [(pad, pad)] + [(0, 0)], 'constant')

            # SimpleITK oracle
            d_sitk = sitk.GetImageFromArray(d, isVector=True)
            d_sitk.SetOrigin(origin)
            d_sitk.SetSpacing(spacing)
            jacobian_sitk = sitk.DisplacementFieldJacobianDeterminant(d_sitk)
            jacobian_sitk = sitk.GetArrayFromImage(jacobian_sitk)

            # Compute Jacobian
            jacobian = pydeform.jacobian(d_sitk)
            jacobian = sitk.GetArrayFromImage(jacobian)

            np.testing.assert_almost_equal(jacobian, jacobian_sitk, decimal=2,
                                           err_msg='Mismatch between `jacobian` and sitk, seed %d' % seed)


    def test_regularize(self):
        df = sitk.GetImageFromArray(rand(10,10,10,3).astype(np.float32), isVector=True)
        full_mask = sitk.GetImageFromArray(np.ones((10,10,10)).astype(np.uint8))

        out = pydeform.regularize(df)
        # Should not be identical
        self.assertFalse(np.array_equal(np.array(out) , np.array(df)))

        # Should fully replicate constraint values
        constraints = sitk.GetImageFromArray(rand(10,10,10,3).astype(np.float32))
        out = pydeform.regularize(df, constraint_mask=full_mask, constraint_values=constraints)
        np.testing.assert_equal(sitk.GetArrayFromImage(out), sitk.GetArrayFromImage(constraints))

    def test_affine(self):
        # Test affine initialization

        affine_transform = sitk.AffineTransform(3)
        affine_transform.SetTranslation((10,10,10))
        affine_transform.SetMatrix((
            2, 0, 0,
            0, 3, 0,
            0, 0, 4
        ))

        # Do a registration pass without actual iterations to see if affine transform is
        # applied to the resulting displacement field
        settings = {
            'max_iteration_count': 0
        }

        fixed = sitk.GetImageFromArray(np.zeros((10,10,10), dtype=np.float32))
        moving = sitk.GetImageFromArray(np.zeros((10,10,10), dtype=np.float32))

        df = pydeform.register(
            fixed,
            moving,
            settings=settings,
            affine_transform=affine_transform
        )

        df = sitk.GetArrayFromImage(df)

        # Ax + b -> A(1, 1, 1) + b -> (2, 3, 4) + (10, 10, 10) -> (12, 13, 14)
        # u(x) = Ax + b - x
        self.assertEqual(df[1,1,1,0], 11)
        self.assertEqual(df[1,1,1,1], 12)
        self.assertEqual(df[1,1,1,2], 13)

if __name__ == '__main__':
    unittest.main()




import multiprocessing as mp
import pydeform


# NOTE: spawn a fresh interpreter for the subprocess, since running
# OpenMP code after a `fork()` syscall may cause a deadlock if OpenMP
# was also used before forking.
ctx = mp.get_context('spawn')


def _registration_worker(q, args, kwargs):
    R""" To be ran in a subprocess.

    Parameters
    ----------
    q: multiprocessing.Queue
        Queue to return the result.
    kwargs: dict
        Keyword arguments for the registration.
    """
    try:
        result = pydeform.register(*args, **kwargs)
    except BaseException as e:
        result = e
    q.put(result)


def register(*args, **kwargs):
    R""" Interruptible version of :func:`pydeform.register`.

    .. note::
        This function calls the registration routine in a subprocess
        in order to handle keyboard interrupts. This has a memory overhead,
        since a new instance of the intepreter is spawned and input objects
        are copied in the subprocess memory.

    .. seealso::
        :func:`pydeform.register`
    """

    # Run call in a subprocess, to handle keyboard interrupts
    q = ctx.Queue()
    p = ctx.Process(target=_registration_worker, args=[q, args, kwargs], daemon=True)
    p.start()
    try:
        result = q.get()
        if isinstance(result, BaseException):
            raise result
        p.join()
    except BaseException as e:
        p.terminate()
        p.join()
        raise e

    return result




import numpy as np
import SimpleITK as sitk
import pydeform
import stk
from . import interruptible

from pydeform import (
    __version__,
    version,
    has_gpu,
    LogLevel
)

def _convert_image(sitk_image):
    """ Converts a SimpleITK.Image to a pydeform.Volume 
    
    Return:
        pydeform.Volume if sitk_image is valid, otherwise None
    """

    if sitk_image is None:
        return None

    return pydeform.Volume(
        sitk.GetArrayViewFromImage(sitk_image),
        sitk_image.GetOrigin(),
        sitk_image.GetSpacing(),
        np.array(sitk_image.GetDirection()).reshape((3,3))
    )

def _convert_transform(transform):
    """ Converts a SimpleITK.AffineTransform to a pydeform.AffineTransform """
    
    translation = np.array(transform.GetTranslation())
    matrix = np.array(transform.GetMatrix()).reshape((3,3))

    # We need to include the fixed parameter (or center) in the offset
    center = np.array(transform.GetCenter())
    offset = translation + center - matrix.dot(center)
    return pydeform.AffineTransform(matrix, offset)


def register(
        fixed_images,
        moving_images,
        *,
        fixed_mask=None,
        moving_mask=None,
        fixed_landmarks=None,
        moving_landmarks=None,
        initial_displacement=None,
        affine_transform=None,
        constraint_mask=None,
        constraint_values=None,
        regularization_map=None,
        settings=None,
        log=None,
        log_level=pydeform.LogLevel.Info,
        silent=True,
        num_threads=0,
        subprocess=False,
        use_gpu=False
        ):
    R""" Perform deformable registration.

    Parameters
    ----------
    fixed_images: Union[SimpleITK.Image, List[SimpleITK.Image]]
        Fixed image, or list of fixed images.

    moving_images: Union[SimpleITK.Image, List[SimpleITK.Image]]
        Moving image, or list of moving images.

    fixed_mask: np.ndarray
        Fixed mask.

    moving_mask: np.ndarray
        Moving mask.

    fixed_landmarks: np.ndarray
        Array of shape :math:`n \times 3`, containing
        `n` landmarks with (x, y, z) coordinates in
        image space. Requires to provide `moving_landmarks`.

    moving_landmarks: np.ndarray
        Array of shape :math:`n \times 3`, containing
        `n` landmarks with (x, y, z) coordinates in
        image space. Requires to provide `fixed_landmarks`.

    initial_displacement: SimpleITK.Image
        Initial guess of the displacement field.

    affine_transform: AffineTransform
        Optional initial affine transformation

    constraint_mask: SimpleITK.Image
        Boolean mask for the constraints on the displacement.
        Requires to provide `constraint_values`.

    constraint_values: SimpleITK.Image
        Value for the constraints on the displacement.
        Requires to provide `constraint_mask`.

    settings: dict
        Python dictionary containing the settings for the
        registration.

    log: Union[StringIO, str]
        Output for the log, either a StringIO or a filename.

    log_level: pydeform.LogLevel
        Minimum level for log messages to be reported.

    silent: bool
        If `True`, do not write output to screen.

    num_threads: int
        Number of OpenMP threads to be used. If zero, the
        number is selected automatically.

    subprocess: bool
        If `True`, run the call in a subprocess and handle
        keyboard interrupts. This has a memory overhead, since
        a new instance of the intepreter is spawned and
        input objects are copied in the subprocess memory.

    use_gpu: bool
        If `True`, use GPU acceleration from a CUDA device.
        Requires a build with CUDA support.

    Returns
    -------
    SimpleITK.Image
        Vector field containing the displacement that
        warps the moving image(s) toward the fixed image(s).
        The displacement is defined in the reference coordinates
        of the fixed image(s), and each voxel contains the
        displacement that allows to resample the voxel from the
        moving image(s).
    """

    if not isinstance(fixed_images, (list, tuple)):
        fixed_images = [fixed_images]
    if not isinstance(moving_images, (list, tuple)):
        moving_images = [moving_images]

    # Get numpy view of the input
    fixed_images = [_convert_image(img) for img in fixed_images]
    moving_images = [_convert_image(img) for img in moving_images]

    if None in fixed_images or None in moving_images:
        raise RuntimeError('Cannot pass None as fixed or moving image')


    # A bit of magic since we can't pass None for arguments expecting stk.Volume
    kwargs = {}
    if initial_displacement:
        kwargs['initial_displacement'] = _convert_image(initial_displacement)
    if affine_transform:
        if (not isinstance(affine_transform, sitk.AffineTransform) or 
            affine_transform.GetDimension() != 3):
            raise ValueError(
                'Expected affine transform to be a 3D SimpleITK.AffineTransform'
            )
        kwargs['affine_transform'] = _convert_transform(affine_transform)
    if constraint_mask:
        kwargs['constraint_mask'] = _convert_image(constraint_mask)
    if constraint_values:
        kwargs['constraint_values'] = _convert_image(constraint_values)
    if regularization_map:
        kwargs['regularization_map'] = _convert_image(regularization_map)
    if fixed_mask:
        kwargs['fixed_mask'] = _convert_image(fixed_mask)
    if moving_mask:
        kwargs['moving_mask'] = _convert_image(moving_mask)

    register = interruptible.register if subprocess else pydeform.register

    # Perform registration through the numpy API
    displacement = register(fixed_images=fixed_images,
                            moving_images=moving_images,
                            settings=settings,
                            log=log,
                            log_level=log_level,
                            silent=silent,
                            num_threads=num_threads,
                            use_gpu=use_gpu,
                            **kwargs
                            )

    # Convert the result to SimpleITK
    out = sitk.GetImageFromArray(np.array(displacement, copy=False), isVector=True)
    out.SetOrigin(displacement.origin)
    out.SetSpacing(displacement.spacing)
    out.SetDirection(displacement.direction.astype(np.float64).flatten())

    return out

def transform(image, df, interp=sitk.sitkLinear):
    R""" Resample an image with a given displacement field.
    
    Parameters
    ----------
    image: SimpleITK.Image
        Image to resample.
        
    df:
        Displacementfield, in reference space coordinates, to apply
    
    interp: pydeform.Interpolator
        Interpolator used in the resampling. E.g. SimpleITK.sitkLinear

    Returns
    -------
    SimpleITK.Image
        Resampled image
    """
    t = sitk.DisplacementFieldTransform(sitk.Cast(df, sitk.sitkVectorFloat64))
    return sitk.Resample(image, df, t, interp)
    

def jacobian(image):
    R""" Compute the Jacobian determinant of a 3D 3-vector field.

    The Jacobian determinant of a vector field

    .. math::
        f(\boldsymbol{x}) =
            (f_1(\boldsymbol{x}),
             f_2(\boldsymbol{x}),
             f_3(\boldsymbol{x}))

    with :math:`\boldsymbol{x} = (x_1, x_2, x_3)`
    is defined as

    .. math::
        J[f] (\boldsymbol{x}) =
        \det \left(
            \frac{\partial f_i}{\partial x_j} (\boldsymbol{x})
        \right)_{ij}

    Parameters
    ----------
    image: SimpleITK.Image
        Input image, should be a 3D 3-vector field.

    Returns
    -------
    SimpleITK.Image
        Scalar image representing the Jacobian determinant.
    """
    result = pydeform.jacobian(_convert_image(image))
    result = sitk.GetImageFromArray(np.array(result, copy=False))
    result.CopyInformation(image)
    return result

def regularize(
    displacement,
    precision = 0.5,
    pyramid_levels = 6,
    constraint_mask = None,
    constraint_values = None):
    """Regularize a given displacement field.

    Parameters
    ----------
    displacement: SimpleITK.Image
        Displacement field used to resample the image.
    precision: float
        Amount of precision.
    pyramid_levels: int
        Number of levels for the resolution pyramid
    constraint_mask: SimpleITK.Image
        Mask for constraining displacements in a specific area, i.e., restricting
        any changes within the region.
    constraint_values: SimpleITK.Image
        Vector field specifying the displacements within the constrained regions.

    Returns
    -------
    SimpleITK.Image
        Scalar volume image containing the resulting displacement field.
    """

    if displacement is None:
        raise ValueError('Expected a displacement field')

    displacement = _convert_image(displacement)

    kwargs = {}
    if constraint_mask:
        kwargs['constraint_mask'] = _convert_image(constraint_mask)
    if constraint_values:
        kwargs['constraint_values'] = _convert_image(constraint_values)

    displacement = pydeform.regularize(
        displacement,
        precision,
        pyramid_levels,
        **kwargs
    )

    # Convert the result to SimpleITK
    out = sitk.GetImageFromArray(np.array(displacement, copy=False), isVector=True)
    out.SetOrigin(displacement.origin)
    out.SetSpacing(displacement.spacing)
    out.SetDirection(displacement.direction.astype(np.float64).flatten())
    return out

def divergence(image):
    R""" Compute the divergence of a 3D 3-vector field.

    The divergence of a vector field

    .. math::
        f(\boldsymbol{x}) =
            (f_1(\boldsymbol{x}),
             f_2(\boldsymbol{x}),
             f_3(\boldsymbol{x}))

    with :math:`\boldsymbol{x} = (x_1, x_2, x_3)`
    is defined as

    .. math::
        \nabla \cdot f (\boldsymbol{x}) =
        \sum_{i=1}^3
            \frac{\partial f_i}{\partial x_i} (\boldsymbol{x})

    Parameters
    ----------
    image: SimpleITK.Image
        Input image, should be a 3D 3-vector field.

    Returns
    -------
    SimpleITK.Image
        Scalar image representing the divergence.
    """
    result = stk.divergence(_convert_image(image))
    result = sitk.GetImageFromArray(np.array(result, copy=False))
    result.CopyInformation(image)
    return result


def rotor(image):
    R""" Compute the rotor of a vector field.

    The rotor of a 3D 3-vector field

    .. math::
        f(\boldsymbol{x}) =
            (f_1(\boldsymbol{x}),
             f_2(\boldsymbol{x}),
             f_3(\boldsymbol{x}))

    with :math:`\boldsymbol{x} = (x_1, x_2, x_3)`
    is defined as

    .. math::
        \nabla \times f(\boldsymbol{x}) =
        \left(
            \frac{\partial f_3}{\partial x_2} -
            \frac{\partial f_2}{\partial x_3},
            \frac{\partial f_1}{\partial x_3} -
            \frac{\partial f_3}{\partial x_1},
            \frac{\partial f_2}{\partial x_1} -
            \frac{\partial f_1}{\partial x_2}
        \right)

    Parameters
    ----------
    image: SimpleITK.Image
        Input image, should be a 3D 3-vector field.

    Returns
    -------
    SimpleITK.Image
        Vector image representing the rotor.
    """
    result = stk.rotor(_convert_image(image))
    result = sitk.GetImageFromArray(np.array(result, copy=False), isVector=True)
    result.CopyInformation(image)
    return result


def circulation_density(image):
    R""" Compute the circulation density of a vector field.

    The circulation density for a 3D 3-vector field is defined as the
    norm of the rotor.

    Parameters
    ----------
    image: SimpleITK.Image
        Input image, should be a 3D 3-vector field.

    Returns
    -------
    SimpleITK.Image
        Vector image representing the circulation density.
    """
    result = stk.circulation_density(_convert_image(image))
    result = sitk.GetImageFromArray(np.array(result, copy=False), isVector=True)
    result.CopyInformation(image)
    return result




import os
import sys
import platform
import subprocess

from pprint import pprint

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

# Parse command line flags
flags = {k: 'OFF' for k in ['--debug', '--use-cuda', '--use-ispc', '--use-itk']}
for flag in flags.keys():
    if flag in sys.argv:
        flags[flag] = 'ON'
        sys.argv.remove(flag)

# Command line flags forwarded to CMake
cmake_cmd_args = []
for f in sys.argv:
    if f.startswith('-D'):
        cmake_cmd_args.append(f)

for f in cmake_cmd_args:
    sys.argv.remove(f)


class CMakeExtension(Extension):
    def __init__(self, name, cmake_lists_dir=''):
        Extension.__init__(self, name, sources=[])
        self.cmake_lists_dir = os.path.abspath(cmake_lists_dir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError('Cannot find CMake executable')

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):

        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cfg = 'Debug' if flags['--debug'] == 'ON' else 'Release'
        build_args = ['--config', cfg]

        cmake_args = [
            '-DDF_BUILD_PYTHON_WRAPPER=ON',
            '-DDF_BUILD_EXECUTABLE=OFF',
            '-DDF_BUILD_TESTS=OFF',
            '-DDF_BUILD_WITH_DEBUG_INFO=%s' % ('ON' if cfg == 'Debug' else 'OFF'),
            '-DCMAKE_BUILD_TYPE=%s' % cfg,
            '-DDF_USE_CUDA=%s' % flags['--use-cuda'],
            '-DDF_USE_ISPC=%s' % flags['--use-ispc'],
            '-DDF_ITK_BRIDGE=%s' % flags['--use-itk'],
            '-DPYTHON_EXECUTABLE=' + sys.executable,
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
        ]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]

        cmake_args += cmake_cmd_args
        pprint(cmake_args)

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(['cmake', ext.cmake_lists_dir] + cmake_args, cwd=self.build_temp)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


setup(
    name='pydeform',
    version='0.5.1',
    author='Simon Ekström',
    author_email='',
    description='',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=['python-stk', 'numpy', 'PyYAML'],
    packages=['pydeform'],
    ext_modules=[CMakeExtension('_pydeform', '.')],
    cmdclass={'build_ext': CMakeBuild},
    zip_safe=False,
)

